-- Auther: aramsay
-- Create Date: 3/22/2009 9:10 PM
-- Status: Alpha
-- TODO:
--      (D) stacking bar frame zorder
--      (D) selecting an existing bar doesn't load values, seen from presets, seems to work with normal bars
--      (D) position method isn't changing right
--      (D) baseCD adjustment doesn't work right
--      (D) width changing messes up short bars
--      (D) loading presets should populate the Update drop down
--      (D) changing a profile doesn't repopulate values list
--
--      (1) presets for every class
--      (1) "energy" bar regen rate scaling for combat and when AR is up
--      (1) "totem" bar type
--      (1) "cast" bar type, don't specify a spell, just a regular cast bar w/ correct timing
--
--      (2) "rune" bar type
--      (2) "internal CD" bars
--      (2) export/import for sharing (can currently use macros as a work-around import)
--
--      (3) reuse bar frames
--      (3) "buff/debuff type" bar to check for buffs in same category (allow 2ndary,etc spells to be specified);
--      (3) localization
--      (3) order the Update drop down to have <Create New> always on top
--
--      (4) multiple base frames
--      (4) "special" bar type(s), for Overpower/Taste For Blood, and other special abilities (use plugins maybe?)
--      (4) detect spec change and load different profile/bars
--      (4) vertical marking (like GCD, energy abilities, or cast times)
--      (4) dot ticks
--      (4) change bars for shape shift forms


AweInspired = LibStub("AceAddon-3.0"):NewAddon("AweInspired", "AceConsole-3.0", "AceEvent-3.0")

local media = LibStub:GetLibrary("LibSharedMedia-3.0")
local maxFramePosition = -1;
local stophide = false;
local bolMoveable = false;
local threshold = 1.5;
local numReset = 30;
local curSelectedBar = "createnew";
local curSelectedPreset = "";
local curSection = "both";

local defaultBarSettings= {
	name = "name",
	spellID = 1,
	barType = "cooldown",
	color = {r=1, g=1, b=1, a=1},
	icon = true,
	frameOffset = "next",
	stackNum = 1,
	longBar = "wait",
				--"wait" holds the bar at full until it's remaining time is less than baseCD then starts
				--"scale" scales the bar to empty slower, losing the relative timing
				--"long" extends the bar beyond the width of baseCD and empties at the same rate of all bars.
				--"hybrid" runs scale while time left is > base CD, then refills and runs normal for the rest.
};

local curBarSettings = defaultBarSettings;
--local curBarSettings = defaults.profile.bars['**'];

local playerClass = select(2, UnitClass('player'));

AweInspired.textures = {};
AweInspired.borders = {};


-----
-- Options passed to AceConfig-3.0 to enable slash commands.
-----
local options = {
	name = "AweInspired",
	handler = AweInspired,
	type = 'group',
	args = {
		deleteframe = {
			type = "input",
			name = "Delete Bar",
			desc = "Type in the spell ID of the bar to delete and press enter to remove it.",
			set = "DeleteFrame",
		},
		clearbars = {
			type = 'execute',
			name = "Clear Bars",
			desc = "Delete all bars",
			func = "ClearBars",
		},
		moveable = {
			type = 'toggle',
			name = "Move Bars",
			desc = "Enable bars to be dragged on the screen.",
			get = function() return bolMoveable; end,
			set = function(info, val)
					bolMoveable = val;
					AweInspired.BaseFrame:EnableMouse(bolMoveable);
					if bolMoveable then
						AweInspired.BaseFrame:SetBackdropColor(1, 1, 1, 1);
					else
						AweInspired.BaseFrame:SetBackdropColor(1, 1, 1, 0);
					end
				end,
			tristate = false,
		},
		general = {
			name = "General Options",
			type = 'group',
			args = {
				texture = {
					type = 'select',
					name = "Bar Texture",
					get = function() return AweInspired.db.profile.texture end,
					set = "SetBarTexture",
					values = AweInspired.textures,
				},
				border = {
					type = 'select',
					name = "Border Texture",
					get = function() return AweInspired.db.profile.border end,
					set = "SetBorderTexture",
					values = AweInspired.borders,
				},
				barheight = {
					type = "range",
					name = "Bar Height",
					desc = "The height of each bar.",
					min = 5,
					max = 100,
					step = 1,
					bigStep = 1,
					get = function(info) return AweInspired.db.profile.fHeight end,
					set = "AdjustHeight",
					isPercent = false,
				},
				barwidth = {
					type = "range",
					name = "Frame Width",
					desc = "Sets the width (in pixels) of bars with a cooldown equal to Base Cooldown (changing Long Bar settings may result in a wider frame).",
					min = 50,
					max = 1000,
					step = 1,
					bigStep = 1,
					get = function(info) return AweInspired.db.profile.fWidth end,
					set = "AdjustWidth",
					isPercent = false,
				},
				basecd = {
					type = "range",
					name = "Base Cooldown",
					desc = "The number in seconds of the max size bar, anything longer than this will be handled based on Long Bar settings.",
					min = 1,
					max = 60,
					step = 0.01,
					bigStep = 1,
					get = function(info) return AweInspired.db.profile.baseCD end,
					set = "AdjustBaseCD",
					isPercent = false,
				},
				scale = {
					type = "range",
					name = "Scale",
					desc = "UI Scale of the bars.",
					min = 0.1,
					max = 2,
					step = 0.01,
					bigStep = 0.05,
					get = function(info) return AweInspired.db.profile.scale; end,
					set = function(info,val)
							AweInspired.BaseFrame:SetScale(val);
							AweInspired.db.profile.scale = val;
						end,
					isPercent = true,
				}
			},
		},
		preset = {
			name = "Load Presets",
			type = 'group',
			args = {
				bar = {
					type = "select",
					name = "Preset",
					order = 10,
					desc = "Select a preset from this list to overwrite your current bars.",
					values = function()
							local retTable = {};
							for i,v in pairs(AweInspired.classDefaultBars) do
								--This Code is just here to catch profiles that aren't complete, I can remove it once I'm done with those
								local nyi = false;
								if v.cooldowns[1] ~= nil then
									if v.cooldowns[1].name == "NYI" then
										nyi = true;
									end
								elseif v.spells[1] ~= nil then
									if v.spells[1].name == "NYI" then
										nyi = true;
									end
								else
									nyi = true;
								end
								if nyi then
									retTable[i] = i.." (NYI - DO NOT USE)";
								else
									retTable[i] = i;
								end
								--End of removable code, when I remove it, uncomment the following line:
								--retTable[i] = i;
							end
							return retTable;
						end,
					set = function(info,val) curSelectedPreset = val; end,
					get = function(info) return curSelectedPreset; end,
					style = "dropdown",
				},
				sections = {
					type = "select",
					name = "Sections",
					order = 15,
					desc = "Select which sections of the preset to load.",
					values = {
						["cooldown"]="Cooldowns",
						["spells"]="Spells",
						["both"]="Both",
					},
					set = function(info,val) curSection = val; end,
					get = function(info) return curSection; end,
					style = "radio",
				},
				apply = {
					type = "execute",
					name = "Apply Preset",
					order = 20,
					desc = "Apply the selected preset bars.",
					func = "LoadPreset",
				},
			},
		},
		newframe = {
			name = "Create Frame",
			type = 'group',
			args = {
				update = {
					type = "select",
					name = "Existing Bar",
					order = 1,
					desc = "Select the name of the existing bar to update.",
					values = {
						["createnew"]="<Create New>",
					},
					set = "UpdateDropDown",
					get = function(info) return curSelectedBar; end,
					style = "dropdown",
				},
				name = {
					type = "input",
					name = "Name",
					order = 5,
					desc = "Name for the bar, only used internally or to delete the bar later",
					set = function(info,val)
							local duplicate = false;
							for i,v in pairs(AweInspired.db.profile.bars) do
								if v.name == val then
									duplicate = true;
								end
							end
							if not duplicate then
								curBarSettings.name = val;
							else
								AweInspired:Print("Name already in use, please use a different name.  To edit existing bar, select it from the dropdown to the left");
							end
						end,
					get = function(info) return curBarSettings.name; end,
				},
				spellID = {
					type = "input",
					name = "Spell ID",
					order = 10,
					desc = "Spell ID or Item ID for new frame.",
					set = function(info,val) curBarSettings.spellID = val; end,
					get = function(info) return curBarSettings.spellID; end,
				},
				bartype = {
					type = "select",
					name = "Bar Type",
					order = 20,
					desc = "Type of bar",
					values = {
						["cooldown"]="Cooldown",
						["buff"]="Buff",
						["debuff"]="Debuff",
						["gcd"]="GCD",
						["combo"]="Combo Points",
						["energy"]="Energy (experimental)",
						["power"]="Power",
						["item"]="Item",
						["stack"]="Stack",
					},
					set = "BartypeDropDown",
					get = function(info) return curBarSettings.barType end,
					style = "dropdown",
				},
				barcolor = {
					type = "color",
					name = "Bar Color",
					order = 30,
					desc = "Color of the bar.",
					set = function(info,r,g,b,a)
							curBarSettings.color = {r=r, g=g, b=b, a=a};
						end,
					get = function(info)
							return curBarSettings.color.r, curBarSettings.color.g, curBarSettings.color.b, curBarSettings.color.a;
						end,
					hasAlpha = true,
				},
				showicon = {
					type = 'toggle',
					name = "Show Icon",
					order = 35,
					desc = "Sets whether or not the bar should display the spell icon.",
					get = function() return curBarSettings.icon; end,
					set = function(info, val) curBarSettings.icon = val; end,
					tristate = false,
				},
				fullstack = {
					type = "input",
					name = "Full Stack Amount",
					order = 40,
					hidden = true,
					desc = "For a \"stack\" type bar, this is how many stacks should be treated as a full bar, for any other bar type, this value is ignored.",
					set = function(info, val) curBarSettings.stackNum = val; end,
					get = function(info) return curBarSettings.stackNum; end,
				},
				longbar = {
					type = "select",
					name = "Long Bar",
					order = 50,
					desc = "When the cooldown is longer than Base Cooldown, how to treat the bar \n\n\"Wait\" - Keep the bar full until the remaining duration is less than Base Cooldown.\n\n\"Scale\" - Scale the time down to fit in the same width, the relative speed of the bars is no longer constant.\n\n\"Long\" - Extend the width of the bar to represent the entire duration/cooldown of the spell.\n\n\"Hybrid\" - Run the spell as \"Scaled\" until the remaining time is less than Base Cooldown, then refill the bar and behave like \"Wait\" for the remainder of the duration.",
					values = {
						["wait"]="Wait",
						["scale"]="Scale",
						["long"]="Long",
						["hybrid"]="Hybrid",
					},
					get = function(info) return curBarSettings.longBar end,
					set = function(info,val) curBarSettings.longBar = val end,
					style = "dropdown",
				},
				position = {
					type = "input",
					name = "Position",
					order = 53,
					desc = "Sets the location of the bar, 0 is the top bar, 1 the next, etc.\n\nEnter \"next\" to place the bar below the bottom bar.",
					set = function(info, val)
							curBarSettings.frameOffset = val;
							if val == "next" then
								info.options.args.newframe.args.insert.disabled = true;
								curInsertMethod = "insert";
							else
								info.options.args.newframe.args.insert.disabled = false;
							end
						end,
					get = function(info) return curBarSettings.frameOffset; end,
				},
				insert = {
					type = "select",
					name = "Position Method",
					order = 57,
					desc = "Determines how the bar is added, \"Insert\" will move all bars of equal or greater position down one level, \"Stack\" will place this bar on top of the existing bar in that position.",
					values = {
						["insert"]="Insert",
						["stack"]="Stack",
					},
					disabled = true,
					get = function(info) return curInsertMethod; end,
					set = function(info,val) curInsertMethod = val; end,
					style = "radio",
				},
				create = {
					type = "execute",
					name = "Create",
					order = 60,
					desc = "Create the bar with the specified settings.",
					func = "CreateBar",
				},
			},
		},
	},
};

-----
-- Default values for saved variables data, used with AceDB-3.0
-----
local defaults = {
--	global = {
--		createdDefaultProfiles = 0,
--	},
	profile = {
		texture = "Blizzard",
		border = "None",
		fHeight = 25,
		fWidth = 300, --The width of the bar frames (in pixels)
		baseCD = 10, --The width of the bar frames (in seconds), anything longer is a "Long Bar"
		point = "CENTER",
		relativeTo = "UIParent",
		relativePoint = "CENTER",
		xOffset = 0,
		yOffset = 0,
		scale = 1,
		bars = {
			['**'] = {
				name = "name",
				spellID = 1,
				barType = "cooldown",
				color = {r=1, g=1, b=1, a=1},
				icon = true,
				longBar = "wait", --what to do if the bar needs to be longer than baseCD,
						--"wait" holds the bar at full until it's remaining time is less than baseCD then starts
						--"scale" scales the bar to empty slower, losing the relative timing
						--"long" extends the bar beyond the width of baseCD and empties at the same rate of all bars.
						--"hybrid" runs scale while time left is > baseCD, then refills and runs normal for the rest.
				frameOffset = -1,
				stackNum = 1,
			},
		},
	},
};



AweInspired.frames = {};
AweInspired.barBackdrop = {
	--bgFile = "Interface/Tooltips/UI-Tooltip-Background",
	edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
	tile = false, tileSize = 0, edgeSize = 12,
	insets = { left = 2, right = 2, top = 2, bottom = 2 }
};

AweInspired.frameBackdrop = {
	bgFile = "Interface/Tooltips/UI-Tooltip-Background",
	--edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
	tile = false, tileSize = 0, edgeSize = 12,
	insets = { left = 0, right = 0, top = 0, bottom = 0 }
};

AweInspired.BaseFrame = CreateFrame("Frame", "AweInspiredBase", UIParent)


-----
-- Initialization, fires when logging on or reloading UI.
-- Register the slash commands, and the bliz interface options menu using AceConfig
-- Set up the saved variables self.db wiht AceDB-3.0.
-- Create the base frame that will be the anchor for the bar frames.
-- Register shared media
-----
function AweInspired:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("AweInspiredDB", defaults);
	for i,v in pairs(self.db.profile.bars) do
		options.args.newframe.args.update.values[v.name] = v.name;
		if v.frameOffset > maxFramePosition then
			maxFramePosition = v.frameOffset;
		end
	end
	options.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db);
	LibStub("AceConfig-3.0"):RegisterOptionsTable("AweInspired", options);
	media.RegisterCallback(self, "LibSharedMedia_Registered");
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("AweInspired", "AweInspired");
	self:RegisterChatCommand("aweinspired", "ChatCommand");
	self:RegisterChatCommand("ais", "ChatCommand");
	self:Print("AweInspired loaded");
	self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig");
	self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig");
	self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig");

--	if self.db.global.createdDefaultProfiles < profileCounter then
--		self:LoadDefaultProfiles();
--	end
end

-----
-- This function will load the default values from ClassDefaults.lua and build profiles for each one
-----
--function AweInspired:LoadDefaultProfiles()
--	--Store current profile name
--	local currentProfile = self.db:GetCurrentProfile();
--	local renamed = false;
--	for i,v in pairs(self.classDefaultBars) do
--		--If they're using a default profile, back it up
--		if not renamed then
--			renamed, currentProfile = self:BackupProfile(currentProfile, i);
--		end
--
--		--Delete existing default profiles (cooldowns, spells, both)
--		self.db:DeleteProfile("AisDefault - "..i.." - Cooldowns");
--		self.db:DeleteProfile("AisDefault - "..i.." - Spells");
--		self.db:DeleteProfile("AisDefault - "..i.." - Both");
--		--Create cooldowns profile and set the bars
--		self.db:SetProfile("AisDefault - "..i.." - Cooldowns");
--		self.db:CopyProfile(currentProfile);
--		wipe(self.db.profile.bars);
--		self.db.profile.bars = v.cooldowns;
--		--Create spells profile and set the bars
--		self.db:SetProfile("AisDefault - "..i.." - Spells");
--		self.db:CopyProfile(currentProfile);
--		wipe(self.db.profile.bars);
--		self.db.profile.bars = v.spells;
--		--Create new profile for both
--		--Load bars from class defaults
--	end
--end

--function AweInspired:BackupProfile(curProfile, baseName)
--	local rename = false;
--	if curProfile == "AisDefault - "..baseName.." - Cooldowns"
--			or curProfile == "AisDefault - "..baseName.." - Spells"
--			or curProfile == "AisDefault - "..baseName.." - Both" then
--		rename = true;
--	end
--
--	if rename then
--		self.db:SetProfile(curProfile.."_old");
--		self.db:CopyProfile(curProfile)
--		return true, curProfile.."_old";
--	else
--		return false, curProfile;
--	end
--end

-----
-- I have no idea what this does, I copied it from ShockAndAwe to try and incorporate SharedMedia
-----
function AweInspired:LibSharedMedia_Registered()
	for k, v in pairs(media:List("statusbar")) do
		self.textures[v] = v;
	end
	for k, v in pairs(media:List("border")) do
		self.borders[v] = v
	end
end

-----
-- Fired when you enable the mod, register all event handlers here
-----
function AweInspired:OnEnable()
	self:LibSharedMedia_Registered();
	--self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED");
	self:RegisterEvent("PLAYER_REGEN_DISABLED");
	self:RegisterEvent("PLAYER_REGEN_ENABLED");

	self:RefreshConfig();
end

-----
-- Fired when you disable the mod, unregister all event handlers here, so they don't fire while disabled
-----
function AweInspired:OnDisable()
	--self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED");
	self:UnregisterEvent("PLAYER_REGEN_DISABLED");
	self:UnregisterEvent("PLAYER_REGEN_ENABLED");
end


-----
-- This function fires when exiting combat.  Disable all bars.
-----
function AweInspired:PLAYER_REGEN_ENABLED()
	for i,v in pairs(self.frames) do
		v:Hide();
		v.icon:Hide();
	end
end

-----
-- This function fires when entering combat, enable all bars.
-----
function AweInspired:PLAYER_REGEN_DISABLED()
	for i,v in pairs(self.frames) do
		--self:Print("showing index ["..i.."]");
		v:Show();
		v.icon:Show();
	end
end

-----
-- Sets up the settings for the base frame, this will be the anchor for the other bar frames.
-- I'm not 100% certain exactly what all this does, but it should allow me to relocate where the bars end up.
-- The logic was copied from ShockAndAwe.
-----
function AweInspired:SetupBaseFrame()
	self.BaseFrame:SetScale(self.db.profile.scale);
	self.BaseFrame:SetFrameStrata("BACKGROUND");
	self.BaseFrame:SetWidth(self.db.profile.fWidth + self.db.profile.fHeight);
	self.BaseFrame:SetHeight(self.db.profile.fHeight * 7);
	self.BaseFrame:SetBackdrop(self.frameBackdrop);
	self.BaseFrame:SetBackdropColor(1, 1, 1, 0);
	self.BaseFrame:SetMovable(true);
	self.BaseFrame:RegisterForDrag("LeftButton");
	self.BaseFrame:SetPoint(self.db.profile.point, self.db.profile.relativeTo, self.db.profile.relativePoint, self.db.profile.xOffset, self.db.profile.yOffset);
	self.BaseFrame:SetScript("OnDragStart",
		function()
			self.BaseFrame:StartMoving();
		end );
	self.BaseFrame:SetScript("OnDragStop",
		function()
			self.BaseFrame:StopMovingOrSizing();
			self.BaseFrame:SetScript("OnUpdate", nil);
			self:FinishedMoving(self.db.profile, self.BaseFrame);
		end ) ;
	self.BaseFrame:Show()
end


-----
-- Called when done dragging and dropping the base frame, saves the new position to the DB.
-- This was copied directly from ShockAndAwe.
-----
function AweInspired:FinishedMoving(var,frame)
	local point, relativeTo, relativePoint, xOffset, yOffset = frame:GetPoint();
	var.point = point
	var.relativeTo = relativeTo
	var.relativePoint = relativePoint
	var.xOffset = xOffset
	var.yOffset = yOffset
end


-----
-- Creates a new bar frame.  This will be located relative to the base frame and offset determined by
-- the value of the frameOffset variable.  This creates the statusbar and spark for the bar as well as
-- setting the overall size.
--
-- Parameters:
--   barSettings: a table of data about the bar, uses the same format as an element of self.db.profile.bars
--   saveBar: a boolean to determine if the bar should be added to saved variables.
-----
--function AweInspired:CreateBarFrame(name, spellID, barType, frameColor, icon, frameLongBar)
function AweInspired:CreateBarFrame(barSettings, saveBar)
	local name = barSettings.name;
	local spellName, spellIcon;
	local idType = self:GetIDType(barSettings.barType)

	if saveBar then
		tinsert(self.db.profile.bars, barSettings);
	end

	if idType == "spellID" then
		spellName, _, spellIcon, _, _, _, _, _, _ = GetSpellInfo(barSettings.spellID);
	elseif idType == "itemID" then
		spellName, _, _, _, _, _, _, _, _, spellIcon, _ = GetItemInfo(barSettings.spellID)
	else
		spellName = "";
		spellIcon = "";
	end

	--Create the frame and icon frame
	self.frames[name] = CreateFrame("Frame", name, self.BaseFrame);
	self.frames[name].icon = CreateFrame("Frame", name.."Icon", self.BaseFrame);

	--Store some frame specific variables
	self.frames[name].spellID = barSettings.spellID;
	self.frames[name].spellName = spellName;
	self.frames[name].barType = barSettings.barType;
	self.frames[name].longBar = barSettings.longBar;
	self.frames[name].frameOffset = barSettings.frameOffset;
	self.frames[name].lastVal = 0;
	self.frames[name].showIcon = barSettings.icon;
	self.frames[name].numReset = 0;
	if barSettings.barType == "gcd" then
		self.frames[name].Width = ((1.5/self.db.profile.baseCD) * self.db.profile.fWidth) + 6;
		self.frames[name].MaxVal = 1.5;
	else
		self.frames[name].Width = self.db.profile.fWidth;
		self.frames[name].MaxVal = self.db.profile.baseCD;
	end

	if barSettings.barType == "stack" then
		self.frames[name].stackNum = barSettings.stackNum
	else
		self.frames[name].stackNum = 0;
	end

	--Set the values of the base frame
	self.frames[name]:SetFrameStrata("LOW");
	self.frames[name]:SetWidth(self.frames[name].Width);
	self.frames[name]:SetHeight(self.db.profile.fHeight);
	self.frames[name]:ClearAllPoints();
	self.frames[name]:SetPoint("TOPLEFT", self.BaseFrame, "TOPLEFT", self.db.profile.fHeight, barSettings.frameOffset * ((-1 * self.db.profile.fHeight) + 3));
	self.frames[name]:SetBackdrop(self.barBackdrop);
	self.frames[name]:SetBackdropColor(0,0,0,0.2);
	self.frames[name]:SetBackdropBorderColor(1,1,1,1);
	self.frames[name]:EnableMouse(false);

	--Create the status bar and set the values
	if not self.frames[name].statusbar then
		self.frames[name].statusbar = CreateFrame("StatusBar", nil, self.frames[name])
	end
	self.frames[name].statusbar:SetFrameLevel(1);
	self.frames[name].statusbar:ClearAllPoints();
	self.frames[name].statusbar:SetHeight(self.db.profile.fHeight - 6);
	self.frames[name].statusbar:SetWidth(self.frames[name].Width - 6);
	self.frames[name].statusbar:SetPoint("TOPLEFT", self.frames[name], "TOPLEFT", 3, -3);
	--self.frames[name].statusbar:SetStatusBarTexture("Interface/Tooltips/UI-Tooltip-Background");
	self.frames[name].statusbar:SetStatusBarTexture(media:Fetch("statusbar", self.db.profile.texture));
	self.frames[name].statusbar:SetStatusBarColor(barSettings.color.r, barSettings.color.g, barSettings.color.b, barSettings.color.a);
	self.frames[name].statusbar:SetMinMaxValues(0, self.frames[name].MaxVal);
	self.frames[name].statusbar:SetValue(self.db.profile.baseCD);

	--Create the spark and set the values
	if not self.frames[name].spark then
		self.frames[name].spark = self.frames[name].statusbar:CreateTexture(nil, "OVERLAY");
	end
	self.frames[name].spark:SetTexture("Interface\\CastingBar\\UI-CastingBar-Spark");
	self.frames[name].spark:SetWidth(16);
	self.frames[name].spark:SetHeight(self.db.profile.fHeight + 16);
	self.frames[name].spark:SetBlendMode("ADD");
	self.frames[name].spark:SetPoint("CENTER",self.frames[name].statusbar,"LEFT",(self.frames[name].Width - 6), 0);
	self.frames[name].spark:Show();

	--Set the OnUpdate event of this frame to call OnBarUpdate
	self.frames[name]:SetScript("OnUpdate",
		function()
			self:OnBarUpdate(self.frames[name]);
		end );

	--If we're supposed to display an icon, set the values for that
	if barSettings.icon then
		self.frames[name].icon:SetWidth(self.db.profile.fHeight - 3)
		self.frames[name].icon:SetHeight(self.db.profile.fHeight - 3)
		self.frames[name].icon:SetPoint("TOPRIGHT", self.frames[name], "TOPLEFT", 0, -2)
		self.frames[name].icon:SetBackdrop({
			bgFile = spellIcon,
			--edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
			tile = false, tileSize = 0, edgeSize = 12,
			insets = { left = 0, right = 0, top = 0, bottom = 0 } })
		self.frames[name].icon:Show();
	end

	self.frames[name]:Show();
	self.frames[name].spark:Show();
end


-----
-- This function returns the type of ID that is used for each barType, most use Spell ID but some use others.
-----
function AweInspired:GetIDType(barType)
	if barType == "cooldown"
			or barType == "buff"
			or barType == "debuff"
			or barType == "energy"
			or barType == "debufftype"
			or barType == "totem"
			or barType == "icd"
			or barType == "gcd"
			or barType == "stack" then
		return "spellID";
	elseif  barType == "item" then
		return "itemID";
	elseif barType == "rune" then
		return "runeID";
	else
		return "none";
	end
end


-----
-- Called constantly from the OnUpdate event of each bar, this is where the bars actually decrease in value over time.
-- the barFrame parameter should contain the frame that triggered the event, decrement it's value.
-----
function AweInspired:OnBarUpdate(barFrame)
	local totCD, timeLeft, count, expirationTime;

	if barFrame.barType == "cooldown" or barFrame.barType == "gcd" then
		timeLeft, totCD = self:GetCooldownValues(barFrame.spellID, barFrame.spellName, true, barFrame.MaxVal, false);
	elseif barFrame.barType == "buff" then
		timeLeft, totCD = self:GetBuffDebuffValues(true, barFrame.spellName, 0);
	elseif barFrame.barType == "debuff" then
		timeLeft, totCD = self:GetBuffDebuffValues(false, barFrame.spellName, 0);
	elseif barFrame.barType == "energy" then
		timeLeft, totCD = self:GetEnergyValues(barFrame.spellID);
	elseif barFrame.barType == "combo" then
		timeLeft, totCD = self:GetComboValues();
	elseif barFrame.barType == "power" then
		timeLeft, totCD = self:GetPowerValues();
	elseif barFrame.barType == "item" then
		timeLeft, totCD = self:GetCooldownValues(barFrame.spellID, barFrame.spellName, false, barFrame.MaxVal, true);
	elseif barFrame.barType == "stack" then
		timeLeft, totCD = self:GetBuffDebuffValues(true, barFrame.spellName, barFrame.stackNum);
		--If we didn't find the stack as a buff, check it as a debuff on the target
		if totCD == 0 then
			timeLeft, totCD = self:GetBuffDebuffValues(false, barFrame.spellName, barFrame.stackNum);
		end
		--If we still don't have a value, check it as a self debuff (ie. arcane blast)
		if totCD == 0 then
			timeLeft, totCD = self:GetBuffDebuffValues(false, barFrame.spellName, barFrame.stackNum, "player");
		end
	else
		return;
	end

	--If the cooldown is shorter than the bar, shrink it down a bit.
	if totCD > 2 and totCD < barFrame.MaxVal then
		barFrame.Width = (totCD / barFrame.MaxVal) * barFrame.Width;
		barFrame:SetWidth(barFrame.Width);
		barFrame.statusbar:SetWidth(barFrame.Width - 6);
		barFrame.statusbar:SetMinMaxValues(0, totCD);
		barFrame.MaxVal = totCD;
	end

	--If there is more time left on the CD than the max value of the status bar, check the longBar setting to see what to do
	if totCD > barFrame.MaxVal then
		if barFrame.longBar == "wait" then
			if timeLeft > barFrame.MaxVal then
				timeLeft = barFrame.MaxVal;
			end
		elseif barFrame.longBar == "scale" then
			timeLeft = (barFrame.MaxVal * timeLeft) / totCD;
		elseif barFrame.longBar == "long" then
			barFrame.Width = (totCD / barFrame.MaxVal) * barFrame.Width;
			barFrame:SetWidth(barFrame.Width);
			barFrame.statusbar:SetWidth(barFrame.Width - 6);
			barFrame.statusbar:SetMinMaxValues(0, totCD);
			barFrame.MaxVal = totCD;
		elseif barFrame.longBar == "hybrid" then
			if timeLeft > barFrame.MaxVal then
				timeLeft = (barFrame.MaxVal * timeLeft) / totCD;
			end
		else
			--This shouldn't ever happen, but just to prevent errors if it does, treat is like "wait"
			if timeLeft > barFrame.MaxVal then
				timeLeft = barFrame.MaxVal;
			end
		end
	end

	if timeLeft <= 0 and barFrame.lastVal - timeLeft > threshold and barFrame.numReset <= numReset then
		timeLeft = barFrame.lastVal
		barFrame.numReset = barFrame.numReset + 1;
	else
		barFrame.numReset = 0;
	end
	
	local sparkLoc = (timeLeft / barFrame.MaxVal) * (barFrame.Width - 6);

	if sparkLoc > 0 then
		barFrame.spark:Show();
	else
		barFrame.spark:Hide();
	end
	
	barFrame.lastVal = timeLeft;
	barFrame.statusbar:SetValue(timeLeft);
	barFrame.spark:SetPoint("CENTER", barFrame.statusbar, "LEFT", sparkLoc, -1);
end


function AweInspired:GetCooldownValues(spellID, spellName, extendExecute, maxVal, isItem)
	local cdStart, totCD;
	if isItem then
		cdStart, totCD, _ = GetItemCooldown(spellID);
	else
		cdStart, totCD, _ = GetSpellCooldown(spellID);
	end
	local timeLeft = (cdStart + totCD) - GetTime();

	if extendExecute then
		local usable, nomana = IsUsableSpell(spellName);
		if not usable and not nomana and timeLeft < 1.5 then
			timeLeft = maxVal;
		end
	end
	return timeLeft, totCD;
end

-----
-- Get the time left and total time for a buff, debuff or stack.
-- Parameters:
--   buff: true if scanning for self buffs on the player, false if checking debuffs on the target
--   spellName: the name of the buff or debuff being queried
--   stackNum: if this is specified, values are returned based on how many stacks there are relative to this number.
--   unitCodeOptional: if this is not nill, it will be used for the target to be checked for buffs, used to override default behavior
-----
function AweInspired:GetBuffDebuffValues(buff, spellName, stackNum, unitCodeOptional)
	local unitCode, unitFilter, timeLeft;
	if buff then
		unitCode = "player";
		unitFilter = "HELPFUL|PLAYER";
	else
		unitCode = "target";
		unitFilter = "HARMFUL|PLAYER";
	end
	if unitCodeOptional ~= nil then
		unitCode = unitCodeOptional;
	end
	local _, _, _, count, _, totCD, expirationTime, _, _, _, _ = UnitAura(unitCode, spellName, nil, unitFilter);

	if totCD == nil then
		timeLeft = 0;
		totCD = 0
	elseif tonumber(stackNum) <= 0 then
		timeLeft = expirationTime - GetTime();
	else
		timeLeft = count * self.db.profile.baseCD / stackNum
		totCD = self.db.profile.baseCD;
	end

	return timeLeft, totCD;
end

function AweInspired:GetEnergyValues(spellID)
	local spellName, _, spellIcon, spellCost, _, spellPowerType, _, _, _ = GetSpellInfo(spellID);
	local energy = UnitMana("player");
	local powerType, _ = UnitPowerType("player");

	if powerType ~= 3 or spellPowerType ~= 3 then
		return 0, 0;
	end
	local timeLeft = (spellCost - energy) / 10.0;

	return timeLeft, (spellCost / 10.0);
end

function AweInspired:GetComboValues()
	local comboPoints = GetComboPoints("player", "target");
	return comboPoints * self.db.profile.baseCD / 5, self.db.profile.baseCD;
end

function AweInspired:GetPowerValues()
	local maxPower = UnitPowerMax("player");
	local curPower = UnitPower("player");
	return curPower * self.db.profile.baseCD / maxPower, self.db.profile.baseCD;
end


-----
-- This function is called with the /ais deleteframe command, it removes the frame for the given name
-----
function AweInspired:DeleteFrame(info, frameName)
	if frameName == nil then
		frameName = 0;
	end

	for i,v in pairs(self.db.profile.bars) do
		if tostring(v.name) == frameName then
			tremove(self.db.profile.bars, i);
		end
	end
	self:RefreshConfig();
end


-----
-- This function removes all bars from saved variables, then refreshes the frame to remove the actual bar frames
-----
function AweInspired:ClearBars(info)
	wipe(self.db.profile.bars);
	wipe(info.options.args.newframe.args.update.values);
	maxFramePosition = -1;
	info.options.args.newframe.args.update.values = { ["createnew"]="<Create New>" };
	self:RefreshConfig();
end



-----
-- This function handles all chat /commands, if no value is specified it opens the interface options for this mod,
-- otherwise it runs the command.
-----
function AweInspired:ChatCommand(input)
	if not input or input:trim() == "" then
		InterfaceOptionsFrame_OpenToCategory(self.optionsFrame);
	else
		LibStub("AceConfigCmd-3.0").HandleCommand(AweInspired, "ais", "AweInspired", input);
	end
end

-----
-- This is called when selecting an new texture from the options menu, it updates the textures of each bar.
-----
function AweInspired:SetBarTexture(info, value)
	self.db.profile.texture = value;
	local barTexture = media:Fetch("statusbar", value);
	for i,v in pairs(self.frames) do
		v.statusbar:SetStatusBarTexture(barTexture);
	end
end

-----
-- This is called when selecting an new border from the options menu, it updates the border of each bar.
-----
function AweInspired:SetBorderTexture(info, value)
	self.db.profile.border = value;
	local newTexture = media:Fetch("border", value);
	if newTexture then
		self.barBackdrop.edgeFile = newTexture;
		for i,v in pairs(self.frames) do
			v:SetBackdrop(self.barBackdrop);
		end
		self:SetupBaseFrame();
	end
end

-----
-- This function is called whenever a profile changes, any changes to saved variables that need to take place immediately should be in here.
-----
function AweInspired:RefreshConfig()
	self:SetBorderTexture(_, self.db.profile.border);
	--self.barBackdrop.edgeFile = media:Fetch("border",self.db.profile.border);


	--This stops the scripts from running, and turns the frames invisible.
	for i,v in pairs(self.frames) do
		v:SetScript("OnUpdate", nil);
		v:Hide();
		v.icon:Hide();
		v.spark:Hide();
	end

	--Clear out all existing frames
	wipe(self.frames);
	self:SetupBaseFrame();

	--Add any saved frames from the new profile.
	for i,v in pairs(self.db.profile.bars) do
		self:CreateBarFrame(v,false);
	end
end


-----
-- This function is called when you change the Bar Type dropdown.  It will enable/disable different controls based on selection
-----
function AweInspired:BartypeDropDown(info, val)
	--First set the value we selected, like any other control
	curBarSettings.barType = val

	--Next alter the frame to only show controls that apply to this bar type, and set hidden values to their defaults
	if val == "cooldown"  or val == "buff" or val == "debuff" or val == "energy" or val == "item" then
		info.options.args.newframe.args.fullstack.hidden = true;
		info.options.args.newframe.args.spellID.hidden = false;
		info.options.args.newframe.args.longbar.hidden = false;
		info.options.args.newframe.args.showicon.hidden = false;
		curBarSettings.stackNum = 0;
	elseif val == "gcd" or val == "combo" or val == "power" then
		info.options.args.newframe.args.fullstack.hidden = true;
		info.options.args.newframe.args.spellID.hidden = true;
		info.options.args.newframe.args.longbar.hidden = true;
		info.options.args.newframe.args.showicon.hidden = true;
		curBarSettings.stackNum = 0;
		curBarSettings.spellID = self.classDefaultGCD[playerClass];
		curBarSettings.longBar = "wait";
		curBarSettings.icon = false;
	elseif val == "stack" then
		info.options.args.newframe.args.fullstack.hidden = false;
		info.options.args.newframe.args.spellID.hidden = false;
		info.options.args.newframe.args.longbar.hidden = true;
		info.options.args.newframe.args.showicon.hidden = false;
		curBarSettings.longBar = "wait";
	else
		--Just in case, if it's anything else, just display all controls
		info.options.args.newframe.args.fullstack.hidden = false;
		info.options.args.newframe.args.spellID.hidden = false;
		info.options.args.newframe.args.longbar.hidden = false;
		info.options.args.newframe.args.showicon.hidden = false;
	end

	--change the names of controls to be more accurate
	if val == "item" then
		info.options.args.newframe.args.spellID.name = "Item ID";
	else
		info.options.args.newframe.args.spellID.name = "Spell ID";
	end
end


-----
-- This function is called when you select an existing bar on the Create Frame window.  If the value is the default "createnew"
-- it leaves the form alone and lets you add a new bar, if it is one of the existing bars, it will allow you to edit those bars.
-----
function AweInspired:UpdateDropDown(info, val)
	local barsIndex = -1;
	curSelectedBar = val;
	if val == "createnew" then
		info.options.args.newframe.args.create.name = "Create";
		info.options.args.newframe.args.name.hidden = false;
		self:BartypeDropDown(info,"cooldown");
		curBarSettings = {
			name = "name",
			spellID = 1,
			barType = "cooldown",
			color = {r=1, g=1, b=1, a=1},
			icon = true,
			frameOffset = "next",
			stackNum = 1,
			longBar = "wait",
		};
	else
		info.options.args.newframe.args.create.name = "Update";
		info.options.args.newframe.args.name.hidden = true;
		curBarSettings.name = val;

		for i,v in pairs(self.db.profile.bars) do
			if v.name == val then
				self:BartypeDropDown(info,v.barType);
				curBarSettings = {
					name = val,
					spellID = v.spellID,
					barType = v.barType,
					color = v.color,
					icon = v.icon,
					frameOffset = v.frameOffset,
					stackNum = v.stackNum,
					longBar = v.longBar,
				};
			end
		end
	end
end

function AweInspired:UpdatePosition()
	local updateBar;
	local moveBarsUp = true;

	--if the position is specified as "next" just append it to the bottom
	if curBarSettings.frameOffset == "next" then
		curBarSettings.frameOffset = maxFramePosition + 1;
	end
	curBarSettings.frameOffset = tonumber(curBarSettings.frameOffset);
	--Update the number to be the new bottom frame if this one is lower than the previous bottom frame
	if curBarSettings.frameOffset > maxFramePosition then
		maxFramePosition = curBarSettings.frameOffset;
	end
	--If we're inserting, move everything else down
	if curInsertMethod == "insert" then
		for i,v in pairs(self.db.profile.bars) do
			self:Print("curOffset: "..curBarSettings.frameOffset);
			self:Print("v.Offset: "..v.frameOffset);
			self:Print("index: "..i);
			self:Print("v.name: "..v.name);
			if v.frameOffset >= curBarSettings.frameOffset then
				v.frameOffset = v.frameOffset + 1;
				if self.frames[v.name] ~= nil then
					self.frames[v.name]:ClearAllPoints();
					self.frames[v.name]:SetPoint("TOPLEFT", self.BaseFrame, "TOPLEFT", self.db.profile.fHeight, v.frameOffset * ((-1 * self.db.profile.fHeight) + 3));
				end
			end

			if v.name == curBarSettings.name then
				updateBar = v;
			end

			if v.frameOffset > maxFramePosition then
				maxFramePosition = v.frameOffset;
			end
		end
		self:Print("curSelectedBar: "..curSelectedBar);
		--If we're working with an existing bar move the ones below it back up
		if curSelectedBar ~= "createnew" then
			--First, find out if it was stacked by checking for at the same offset
			for i,v in pairs(self.db.profile.bars) do
				if v.frameOffset == updateBar.frameOffset then
					moveBarsUp = false;
				end
			end
			if moveBarsUp then
				--If there weren't any stacked bars we need to move everyone below the current node up one level
				for i,v in pairs(self.db.profile.bars) do
					if v.frameOffset >= updateBar.frameOffset then
						v.frameOffset = v.frameOffset - 1;
						self.frames[v.name]:ClearAllPoints();
						self.frames[v.name]:SetPoint("TOPLEFT", self.BaseFrame, "TOPLEFT", self.db.profile.fHeight, v.frameOffset * ((-1 * self.db.profile.fHeight) + 3));
					end
				end
			end
		end
	end
end

function AweInspired:CreateBar(info)
	local moveBarsUp = true;
	self:UpdatePosition();

	if curSelectedBar ~= "createnew" then
		--updating an existing bar
		for i,v in pairs(self.db.profile.bars) do
			if v.name == curBarSettings.name then
				v.spellID = curBarSettings.spellID;
				v.barType = curBarSettings.barType;
				v.color = curBarSettings.color ;
				v.stackNum = curBarSettings.stackNum ;
				v.longBar = curBarSettings.longBar;
				v.icon = curBarSettings.icon;
				v.frameOffset = curBarSettings.frameOffset;
				self:RefreshConfig();
				curSelectedBar = "createnew";
				self:UpdateDropDown(info, "createnew");
			end
		end
	else
		--creating a new bar
		self:CreateBarFrame({name=curBarSettings.name, spellID=curBarSettings.spellID, barType=curBarSettings.barType, color=curBarSettings.color, icon=curBarSettings.icon, longBar=curBarSettings.longBar, frameOffset=curBarSettings.frameOffset, stackNum=curBarSettings.stackNum}, true);
		--self:CreateBarFrame(curBarSettings, true);
		info.options.args.newframe.args.update.values[curBarSettings.name] = curBarSettings.name;
	end
	curBarSettings = {
		name = "name",
		spellID = 1,
		barType = "cooldown",
		color = {r=1, g=1, b=1, a=1},
		icon = true,
		frameOffset = "next",
		stackNum = 1,
		longBar = "wait",
	};
	self:BartypeDropDown(info, "cooldown");
	--curBarSettings = defaultBarSettings;
	--curBarSettings = defaults.profile.bars['**'];
end


function AweInspired:DeepCopyBars(srcBar)
	return {
		name = srcBar.name, 
		spellID = srcBar.spellID,
		barType = srcBar.barType,
		color = srcBar.color,
		icon = srcBar.icon,
		longBar = srcBar.longBar,
		frameOffset = srcBar.frameOffset,
		stackNum = srcBar.stackNum,
	};
end


-----
-- This function will load a preset bar setup from ClassDefaults.lua based on what selections were made in the frame.
-----
function AweInspired:LoadPreset(info)
	local presetGroup = self.classDefaultBars[curSelectedPreset];
	local maxIndex = 0;
	local maxOffset = 0;

	wipe(self.db.profile.bars);
	maxFramePosition = -1;

	if curSection == "cooldown" or curSection =="both" then
		for i,v in pairs(presetGroup.cooldowns) do
			maxIndex = i;
			if v.frameOffset > maxOffset then
				maxOffset = v.frameOffset;
			end
			info.options.args.newframe.args.update.values[v.name] = v.name;
			self.db.profile.bars[i] = self:DeepCopyBars(v);
			self.db.profile.bars[i].frameOffset = v.frameOffset;
		end
		maxOffset = maxOffset + 1;
	end

	if curSection == "spells" or curSection =="both" then
		for i,v in pairs(presetGroup.spells) do
			info.options.args.newframe.args.update.values[v.name] = v.name;
			self.db.profile.bars[maxIndex + i] = self:DeepCopyBars(v);
			self.db.profile.bars[maxIndex + i].frameOffset = v.frameOffset + maxOffset;
		end
	end

	self:RefreshConfig();
end


-----
-- Change the saved baseCD value, also adjust any values which this affects, for immediate results
-----
function AweInspired:AdjustBaseCD(info,val)
	self.db.profile.baseCD = val;

	for i,v in pairs(self.frames) do
		if v.barType == "gcd" then
			v.Width = ((1.5/val) * self.db.profile.fWidth) + 6;
			v:SetWidth(v.Width);
			v.statusbar:SetWidth(v.Width - 6);
		else
			v.MaxVal = val;
		end
	end
end

-----
-- Change the saved height value, also adjust any values which this affects, for immediate results
-----
function AweInspired:AdjustHeight(info,val)
	self.db.profile.fHeight = val
	self.BaseFrame:SetWidth(self.db.profile.fWidth + val);
	self.BaseFrame:SetHeight(self.db.profile.fHeight * 7);

	for i,v in pairs(self.frames) do
		v:SetHeight(val);
		v:ClearAllPoints();
		v:SetPoint("TOPLEFT", self.BaseFrame, "TOPLEFT", val, v.frameOffset * ((-1 * val) + 3));
		v.statusbar:SetHeight(val - 6);
		v.spark:SetHeight(val + 16);
		if v.showIcon then
			v.icon:SetWidth(val - 3);
			v.icon:SetHeight(val - 3);
		end
	end
end

-----
-- Change the saved width value, also adjust any values which this affects, for immediate results
-----
function AweInspired:AdjustWidth(info,val)
	self.db.profile.fWidth = val
	self.BaseFrame:SetWidth(val + self.db.profile.fHeight);

	for i,v in pairs(self.frames) do
		if v.barType == "gcd" then
			v.Width = ((1.5/self.db.profile.baseCD) * val) + 6;
		else
			v.Width = val;
		end
		v:SetWidth(v.Width);
		v.statusbar:SetWidth(v.Width - 6);
	end
end
